import { Cookies } from 'react-cookie';

export class AuthorizationHandler {
  private static _instance: AuthorizationHandler;
  private accessToken = '';
  private authorized = false;
  private userID = '';
  private sessionID = '';
  private callback?: () => void = undefined;

  // instance gets the AuthorizationHandler singleton instance
  public static instance(): AuthorizationHandler {
    if (!AuthorizationHandler._instance) {
      AuthorizationHandler._instance = new AuthorizationHandler();
    }

    return AuthorizationHandler._instance;
  }

  // SetCallback sets a function that will be called when the auth state changes
  // TODO: This should be done with a store or custom hook
  public setCallback(callback: () => void) {
    this.callback = callback;
  }

  // MARK: Authorization code

  /**
   * Authorize attempts to authorize using a refresh-token set as a cookie. If the user has been inactive for more than 7 days this cookie will be gone.
   * @returns true is authorization was successful, else returns false
   */
  public async Authorize(): Promise<boolean> {
    // Attempt to log in with a refresh-token
    const authResponse = await this.getNewAccessToken();

    // If the request was a success, we have an accessToken, userID and sessionID
    if (authResponse.success) {
      // Store them
      this.userID = authResponse.userID ?? '';
      this.sessionID = authResponse.sessionID ?? '';
      this.SetAccessToken(authResponse.accessToken ?? '');
    }

    return new Promise((resolve) => {
      resolve(authResponse.success);
    });
  }

  /**
   * getNewAccessToken gets a new access token using the refresh-token cookie
   * @returns an authResponse containing details
   */
  private async getNewAccessToken(): Promise<authResponse> {
    // If we have an access token already, append it to the url as a query param to keep sessionID the same
    let url = 'https://datastrophe.science.uu.nl/auth/refresh';
    if (this.accessToken != '') {
      url += '?access_token=' + this.accessToken;
    }

    return new Promise<authResponse>((resolve) => {
      fetch(url, {
        method: 'GET',
        credentials: 'include',
      })
        .then((response) => {
          if (!response.ok) {
            throw Error(response.statusText);
          }
          return response.json();
        })
        .then((responseJSON) => {
          resolve({
            success: true,
            accessToken: responseJSON.accessToken,
            userID: responseJSON.userID,
            sessionID: responseJSON.sessionID,
          });
        })
        .catch(() => {
          // User is not authorized
          resolve({ success: false });
        });
    });
  }

  /**
   * refreshTokens refreshes tokens
   */
  private async refreshTokens() {
    console.log('refreshing tokens');
    // Get a new access + refresh token pair
    const authResponse = await this.getNewAccessToken();

    if (authResponse.success) {
      // Set the new access token
      this.accessToken = authResponse.accessToken ?? '';

      // Initialise the new refresh token
      this.initialiseRefreshToken();
    }
  }

  /**
   * initialiseRefreshToken attempts to initialise a refresh token
   */
  private async initialiseRefreshToken() {
    fetch('https://datastrophe.science.uu.nl/auth/refresh', {
      method: 'POST',
      credentials: 'include',
    })
      .then((response) => {
        if (!response.ok) {
          throw Error(response.statusText);
        }
      })
      .catch((error) => {
        console.error(error);
      });
  }

  // MARK: Getters

  /**
   * Authorized returns the current authorization status
   * @returns true if authorized
   */
  Authorized(): boolean {
    return this.authorized;
  }

  /**
   * AccessToken returns the current access token
   * @returns token
   */
  AccessToken(): string {
    return this.accessToken;
  }

  /**
   * UserID returns the current user' ID
   * @returns id
   */
  UserID(): string {
    return this.userID;
  }

  /**
   * SessionID returns the current session' ID
   * @returns id
   */
  SessionID(): string {
    return this.sessionID;
  }

  // MARK: Setters
  /**
   * SetAccessToken sets the current access token (should only be called by the sign-in component)
   * @param accessToken
   */
  SetAccessToken(accessToken: string) {
    this.accessToken = accessToken;

    console.log(this.accessToken);

    // Activate the refresh token
    this.initialiseRefreshToken();

    // Start the automatic refreshing every 10 minutes
    setInterval(() => {
      this.refreshTokens();
    }, 10 * 60 * 1000);

    // TODO: Change auth state
    if (this.callback) {
      this.callback();
    }
  }
}

type authResponse = {
  success: boolean;
  accessToken?: string;
  userID?: string;
  sessionID?: string;
};
